/***************************************************************************
 *   Copyright (C) 2004-2008 by Pere Constans
 *   constans@molspaces.com
 *   cb2Bib version 0.9.4
 *   See the LICENSE file that comes with this distribution
 ***************************************************************************/
#include "c2bBibPreparser.h"

#include "authorString.h"
#include "c2b.h"
#include "c2bBibParser.h"
#include "c2bCiteIDMaker.h"
#include "c2bClipEdit.h"
#include "c2bPreprocess.h"
#include "c2bSettings.h"
#include "journalDB.h"

#include <QApplication>
#include <QComboBox>
#include <QFile>
#include <QMessageBox>


c2bBibParser::c2bBibParser(QObject *parent) : bibParser(parent)
{
    settings = c2bSettingsP;
    c2bMain = c2b::mainWidget();

    // Creating Journal Name Database
    QString journal_f = settings->fileName("cb2Bib/JournalFile");
    if (journal_f.isEmpty())
        QMessageBox::information(c2bMain, tr("Information - cb2Bib"),
                                 tr("No Abbreviation Journal file has been specified.\n\n"
                                    "Note: Abbreviation files are specified through the cb2Bib Configure dialog."),
                                 QMessageBox::Ok);
    else
    {
        QFile jdbfile(journal_f);
        if (!jdbfile.open(QIODevice::ReadOnly | QIODevice::Text))      // jdbfile.exists wouldn't produce an errorString
            QMessageBox::warning(c2bMain, tr("Warning - cb2Bib"),
                                 tr("Unable to open the Journal Abbreviation file %1 for reading.\nError: '%2'.\n\n"
                                    "Note: Abbreviation files are specified through the cb2Bib Configure dialog.")
                                 .arg(journal_f).arg(jdbfile.errorString()),
                                 QMessageBox::Ok);
        else
            jdbfile.close();
    }
    JDB = new journalDB(journal_f);

    // Creating Preparser object
    prParser = new c2bBibPreparser(this);
    connect(prParser, SIGNAL(preParserDataAvailable(const QString&)), this, SIGNAL(preParserDataAvailable(const QString&)));

    // Creating Stream Preprocess Object
    prProc = new c2bPreprocess(c2bMain);
}

c2bBibParser::~c2bBibParser()
{
    delete JDB;
}


QString c2bBibParser::makeBib(bool sig)
{
    QString BibString;
    if (sig)
        BibString = "% cb2Bib " + C2B_VERSION + "\n";
    else
        BibString = "";
    BibString += "@" + bibTypes->currentText() + "{" + bibFields["id"]->text();
    QStringList::Iterator it = bibFieldList.begin();
    while (it != bibFieldList.end())
    {
        QString fvalue = bibFields[*it]->text();
        if (!fvalue.isEmpty())
        {
            if (settings->value("cb2Bib/ConvertReferenceToLaTeX").toBool())
                c2bUtils::c2bToBib(fvalue);
            QString fd = *it;
            QString dum;
            dum.fill(' ', 12 - fd.length());
            if (*it == "month" && settings->value("cb2Bib/PostprocessMonth").toBool())
                BibString += ",\n" + fd + dum + " =   " + fvalue;
            else
                BibString += ",\n" + fd + dum + " = {" + fvalue + "}";
        }
        it++;
    }
    BibString += "\n}\n";
    if (sig)
        BibString += "\n";
    return BibString;
}

void c2bBibParser::makeCiteID()
{
    citeIDM->makeCiteID();
}

QString c2bBibParser::excerpt(const QString& text, const QStringList& hints)
{
    QString txt = removeTags(text);
    txt.replace(QRegExp("(http://|https://|ftp://|www\\.|ftp\\.)(www\\.|ftp\\.){0,1}\\S+"), " ");
    txt.replace(QRegExp("\\b[A-Z]+\\b"), " ");
    txt.replace(QRegExp("\\d"), " ");
    txt.replace(QRegExp("\\W"), " ");
    txt.replace(QRegExp("\\b\\w{1,2}\\b"), " ");
    txt.replace(QRegExp("\\b(about|and|are|com|for|from|how|into|that|the|their|this|where|with|www)\\b", Qt::CaseInsensitive), " ");
    txt.replace(QRegExp("\\b(january|february|march|april|may|june|july|august|september|october|november|december)\\b",
                        Qt::CaseInsensitive), " ");
    txt = hints.join(" ") + " " + txt;
    txt = txt.simplified();
    QStringList txtlist = txt.split(" ", QString::SkipEmptyParts);
    QStringList txtlistSimp;
    for (int i = 0; i < qMin(15, txtlist.count()); ++i)
        txtlistSimp.append(txtlist.at(i));
    return txtlistSimp.join(" ");
}

QString c2bBibParser::PProc(QString field, QString value)
{
    // Post Processing of Fields
    value = removeTags(value);
    if (value.isEmpty())
        return (value);
    c2bUtils::fullBibToC2b(value);

    bool fullNames = settings->value("cb2Bib/UseFullNames").toBool();
    bool titleDoubleBraces = settings->value("cb2Bib/UseDoubleBraces").toBool();
    if (field == "author")
    {
        c2bUtils::debug(value);
        authorString AuthorList = authorString(value, fullNames);
        value = AuthorList.bib();
    }
    else if (field == "addauthors")
    {
        QString PAuthors = bibFields["author"]->text();
        c2bUtils::debug(value);
        authorString AuthorList = authorString(value, fullNames);
        if (PAuthors.isEmpty())
            value = AuthorList.bib();
        else
            value = PAuthors + " and " + AuthorList.bib();
    }
    else if (field == "editor")
    {
        c2bUtils::debug(value);
        authorString EditorList = authorString(value.remove(QRegExp("\\((Editor|Editors|Ed|Eds)\\)", Qt::CaseInsensitive)), fullNames);
        value = EditorList.bib();
    }
    else if (field == "addeditor")
    {
        QString PEditor = bibFields["editor"]->text();
        c2bUtils::debug(value);
        authorString EditorList = authorString(value.remove(QRegExp("\\((Editor|Editors|Ed|Eds)\\)", Qt::CaseInsensitive)), fullNames);
        if (PEditor.isEmpty())
            value = EditorList.bib();
        else
            value = PEditor + " and " + EditorList.bib();
    }
    else if (field == "journal")
    {
        QString jvalue;
        if (settings->value("cb2Bib/SetJournalsToFullname").toBool())
        {
            value = JDB->retrieveFull(value);
            jvalue = JDB->retrieve(value);
        }
        else
        {
            value = JDB->retrieve(value);
            jvalue = JDB->retrieveFull(value);
        }
    }
    else if (field == "month")
    {
        if (settings->value("cb2Bib/PostprocessMonths").toBool())
            value = MDB->retrieve(value);
    }
    else if (field == "pages")
        value = setPages(value);
    else if (field == "title" || field == "booktitle")
    {
        if (titleDoubleBraces)
            if (!value.contains(QRegExp("^\\{.+\\}$")))
                value = "{" + value + "}";
        if (c2bUtils::isUpperCaseString(value))
            value = c2bUtils::setCapitalization(value);
    }
    else if (field == "addtitle")
    {
        if (titleDoubleBraces)
            if (!value.contains(QRegExp("^\\{.+\\}$")))
                value = "{" + value + "}";
        if (c2bUtils::isUpperCaseString(value))
            value = c2bUtils::setCapitalization(value);
        QString PTitle = bibFields["title"]->text();
        if (!PTitle.isEmpty())
            value = PTitle + ": " + value;
    }
    return value.simplified();
}

QString c2bBibParser::abbreviatedJournal(const QString& j)
{
    return (JDB->retrieve(j));
}

QString c2bBibParser::fullJournal(const QString& j)
{
    return (JDB->retrieveFull(j));
}

void c2bBibParser::setJournal()
{
    bibFields["journal"]->setText(PProc("journal", bibFields["journal"]->text()));
}

QString c2bBibParser::setJournalsToFull(const QString& text)
{
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    QString substituted_text = text;
    QRegExp jnre("\\bjournal\\s*=\\s*[\\{\"](.*)[\\}\"]", Qt::CaseInsensitive);
    jnre.setMinimal(true);
    int pos = 0;
    uint nj = 0;
    while (pos >= 0)
    {
        pos = jnre.indexIn(substituted_text, pos);
        if (pos > -1)
        {
            QString line = jnre.cap(0);
            QString jn = jnre.cap(1);
            line.replace(jn, fullJournal(jn));
            substituted_text.replace(pos, jnre.matchedLength(), line);
            pos += line.length();
            nj++;
        }
        c2b::showMessage(tr("Processed %1 journal names...").arg(nj));
        QCoreApplication::processEvents();
    }
    QApplication::restoreOverrideCursor();
    c2b::showMessage(tr("Processed %1 journal names.").arg(nj));
    return (substituted_text);
}

QString c2bBibParser::setJournalsToAbbreviated(const QString& text)
{
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    QString substituted_text = text;
    QRegExp jnre("\\bjournal\\s*=\\s*[\\{\"](.*)[\\}\"]", Qt::CaseInsensitive);
    jnre.setMinimal(true);
    int pos = 0;
    uint nj = 0;
    while (pos >= 0)
    {
        pos = jnre.indexIn(substituted_text, pos);
        if (pos > -1)
        {
            QString line = jnre.cap(0);
            QString jn = jnre.cap(1);
            line.replace(jn, abbreviatedJournal(jn));
            substituted_text.replace(pos, jnre.matchedLength(), line);
            pos += line.length();
            nj++;
        }
        c2b::showMessage(tr("Processed %1 journal names...").arg(nj));
        QCoreApplication::processEvents();
    }
    QApplication::restoreOverrideCursor();
    c2b::showMessage(tr("Processed %1 journal names.").arg(nj));
    return (substituted_text);
}

void c2bBibParser::setRefType(QString type)
{
    //  Post Processing of Reference Types
    int i = 0;
    while (i < bibTypes->count())
    {
        if (bibTypes->itemText(i) == type)
        {
            bibTypes->setCurrentIndex(i);
            return;
        }
        i++;
    }
    bibTypes->setCurrentIndex(0);
    bibTypes->setItemText(0, type);
}

void c2bBibParser::clearFields()
{
    QStringList::Iterator it = bibFieldList.begin();
    while (it != bibFieldList.end())
    {
        bibFields[*it]->clear();
        it++;
    }
    bibTypes->setCurrentIndex(1);
    m_auto_regc = false;
}



/****************************************************************************

AUTOMATIC BIB CAPTION

*****************************************************************************/

void c2bBibParser::preParseItem(const QString& text)
{
    prParser->preParseItem(text);
}

/*! \page bibproc Extracting Data from the Clipboard

    Clipboard contents is processed according to the following rules:

    - Perform external, user-defined preparsing on input stream. See \ref c2bconf_input.

    - Perform user-defined substitutions on input stream. See \ref c2bconf_input.

    - Check if input stream is already a BibTeX entry. If so, process entry.

    - Check if input stream is a PubMed - Medline Journal entry. If so, process entry.

    - Preprocess author names: PI JOAN III -> Pi III, J.
    (care of name prefixes, suffixes, and removal of ambiguities).


    If otherwise,

    - Extract DOI \n (DOI, URL and FILE/PDF are preprocessed, performed before
    the automatic recognition takes place.)

    - Extract URL

    - Remove leading and trailing white spaces, TABs and CRs.

    - "\r\n", "\n" and/or "\r" replaced by the line indicator tag "<NewLineN>".

    - Replace "\t" and five or more consecutive "\s" by the tabular tag "<TabN>".

    - Simplify White Spaces

    - Start the automatic recognition engine.


    If the automatic recognition engine fails, optionally, a heuristic guessing
    will be performed. See \ref heuristic_guess.

*/
void c2bBibParser::processItem(QString text)
{
    prProc->preprocStream(text);

    if (hasBibTeX(text))
    {
        bibEditor->setText(text);
        const bibReference bRef = wholeReference(text);
        bibReferenceIterator i(bRef);
        while (i.hasNext())
        {
            i.next();
            bibFields[i.key()]->setText(PProc(i.key(), i.value()));
        }
        setRefType(bRef.typeName);
        c2bAutoRecogStr = tr("Processed as 'BibTeX'.");
        m_auto_regc = true;
        c2b::showMessage(c2bAutoRecogStr);
        return;
    }

    QRegExp ml("^PMID\\s*-");
    if (text.contains(ml))
    {
        setRefType("article");
        bibEditor->setText(text);
        text.replace(QRegExp("[\\n\\r]\\s*([A-Z]{1,4}\\s*-)"), "][\\1");
        text = text.simplified();
        if (!text.contains(QRegExp("\\[FAU\\s+-")))
            text.replace(QRegExp("\\[(AU\\s*-\\s*[-'\\w]+)"), "[F\\1 ");
        QStringList fList = text.split("][");
        QString kw = "";
        // Link set according to: "Creating a Web Link to the Entrez Databases",
        // http://www.ncbi.nlm.nih.gov/books/bv.fcgi?rid=helplinks.chapter.linkshelp#linkshelp.Retrieve_PubMed_Cita
        QString pubmed_url = "http://www.ncbi.nlm.nih.gov/pubmed/%1";
        QRegExp fld("^([A-Z]{1,4})\\s{0,1}-\\s*(.+)$");
        fld.setPatternSyntax(QRegExp::RegExp2);
        for (QStringList::Iterator it = fList.begin(); it != fList.end(); ++it)
        {
            if (fld.indexIn(*it) == -1)
                continue;
            QString tag = fld.cap(1);
            QString value = fld.cap(2);
            if (tag == "AB")
                bibFields["abstract"]->setText(PProc("abstract", value));
            else if (tag == "FAU")
                bibFields["author"]->setText(PProc("addauthors", medlToc2b(value)));
            else if (tag == "TA")
                bibFields["journal"]->setText(PProc("journal", value));
            else if (tag == "IP")
                bibFields["number"]->setText(PProc("number", value));
            else if (tag == "PG")
                bibFields["pages"]->setText(PProc("pages", value));
            else if (tag == "TI")
                bibFields["title"]->setText(PProc("title", value));
            else if (tag == "PMID")
                bibFields["url"]->setText(PProc("url", pubmed_url.arg(value)));
            else if (tag == "VI")
                bibFields["volume"]->setText(PProc("volume", value));
            else if (tag == "AID")
            {
                if (value.contains("[doi]"))
                    bibFields["doi"]->setText(PProc("doi", value.remove("[doi]")));
            }
            else if (tag == "DP")
                bibFields["year"]->setText(PProc("year", value.replace(QRegExp("^([\\d\\s]+).*$"), "\\1")));
            else if (tag == "MH")
                kw += "; " + value.trimmed();
        }
        if (!kw.isEmpty())
            bibFields["keywords"]->setText(PProc("keywords", kw.remove(0, 2)));
        c2bAutoRecogStr = QString("Processed as 'PubMed - Medline Journals'.");
        m_auto_regc = true;
        c2b::showMessage(c2bAutoRecogStr);
        return;
    }

    QRegExp rxdoi("(10.[\\d\\.]+/\\S+)");
    int ndoi = rxdoi.indexIn(text);
    if (ndoi > -1)
        bibFields["doi"]->setText(PProc("doi", rxdoi.cap(1)));

    QRegExp rxhtml("((http://|https://|ftp://|www\\.|ftp\\.)(www\\.|ftp\\.){0,1}\\S+)");
    int nhtml = rxhtml.indexIn(text);
    if (nhtml > -1)
        bibFields["url"]->setText(PProc("url", rxhtml.cap(1)));

    QString tagged_text = setTags(text);
    bibEditor->setText(text, tagged_text);

    QString regular_expression_f = settings->fileName("cb2Bib/RegularExpressionFile");
    if (regular_expression_f.isEmpty())
    {
        QMessageBox::information(c2bMain, tr("Information - cb2Bib"),
                                 tr("No RegExp filename has been specified.\n\n"
                                    "Note: RegExp files are specified through the cb2Bib Configure dialog."),
                                 QMessageBox::Ok);
        return;
    }
    QFile file(regular_expression_f);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        QMessageBox::warning(c2bMain, tr("Warning - cb2Bib"),
                             tr("Unable to open the RegExp file %1 for reading.\nError: '%2'.\n\n"
                                "Note: RegExp files are specified through the cb2Bib Configure dialog. "
                                "After manual matching, patterns can be edited and stored in a "
                                "RegExp file for future autodetection.")
                             .arg(regular_expression_f).arg(file.errorString()),
                             QMessageBox::Ok);
        return;
    }

    QString ItemX;
    QString line;
    QString reftype;
    QString fieldset;
    QTextStream stream(&file);
    int nfilts = 0;

    while (!stream.atEnd())
    {
        line = stream.readLine();
        if (!(line.isEmpty() || line.contains(QRegExp("^#"))))
        {
            reftype = stream.readLine();
            fieldset = stream.readLine();
            ItemX = stream.readLine();

            c2bUtils::debug(tr("The RegExp file contains1: |%1|").arg(line));
            c2bUtils::debug(tr("The RegExp file contains2: |%1|").arg(reftype));
            c2bUtils::debug(tr("The RegExp file contains3: |%1|").arg(fieldset));
            c2bUtils::debug(tr("The RegExp file contains4: |%1|").arg(ItemX));

            QRegExp rx(ItemX);
            rx.setMinimal(true);
            if (!rx.isValid())
                qDebug(tr("[cb2Bib] RegExp |%1| is not valid.").arg(ItemX).toLatin1());

            QStringList list = fieldset.split(" ");
            nfilts++;
            int nfields = rx.numCaptures();
            int ncap = rx.indexIn(tagged_text);
            c2bUtils::debug(tr("Readable Fields: |%1|").arg(nfields));
            c2bUtils::debug(tr("Captures: |%1|").arg(ncap));

            if (ncap > -1)
            {
                for (int i = 0; i < nfields; i++)
                {
                    QString listi = list.at(i);
                    int ii = i + 1;
                    c2bUtils::debug(QString("Fields in Template %1: |%2|").arg(i).arg(rx.cap(ii)));
                    if (listi.contains(field_re))
                    {
                        if (listi == "author")
                            // Reminder: "addauthors" requires to init bibFields["author"]
                            bibFields[listi]->setText(PProc("addauthors", rx.cap(ii)));
                        else if (listi == "editor")
                            // Reminder: "addeditor" requires to init bibFields["editor"]
                            bibFields[listi]->setText(PProc("addeditor", rx.cap(ii)));
                        else if (listi == "title")
                            // Reminder: "addtitle" requires to init bibFields["title"]
                            bibFields[listi]->setText(PProc("addtitle", rx.cap(ii)));
                        else
                            bibFields[listi]->setText(PProc(listi, rx.cap(ii)));
                    }
                }
                setRefType(reftype);
                c2bAutoRecogStr = tr("Processed as '%1'.").arg(line);
                m_auto_regc = true;
                c2b::showMessage(c2bAutoRecogStr);
                file.close();
                return;
            }
        }
    }
    file.close();

    // Heuristic Bib Parsing
    if (settings->value("cb2Bib/DoHeuristicGuess").toBool())
    {
        QString clean_text = removeTags(tagged_text);
        guessAbstract(tagged_text);
        guessKeywords(tagged_text);
        guessVolume(clean_text);
        guessNumber(clean_text);
        guessPages(clean_text);
        guessYear(clean_text);
        guessTitle(tagged_text);
        guessISBN(clean_text);
        guessJournal(clean_text);
        c2bAutoRecogStr = tr("Applied %1 filters: No automatic format detection. %2 fields guessed.")
                          .arg(nfilts).arg(countFields());
    }
    else
        c2bAutoRecogStr = tr("Applied %1 filters: No automatic format detection.").arg(nfilts);

    c2b::showMessage(c2bAutoRecogStr);
}

void c2bBibParser::guessItem(const QString& text)
{
    QString clean_text = text.simplified();
    QString tagged_text = setTags(text);
    guessAbstract(tagged_text);
    guessKeywords(tagged_text);
    guessVolume(clean_text);
    guessNumber(clean_text);
    guessPages(clean_text);
    guessYear(clean_text);
    guessTitle(tagged_text);
    guessISBN(clean_text);
    guessJournal(clean_text);
    c2bAutoRecogStr = tr("%2 fields guessed.").arg(countFields());
    c2b::showMessage(c2bAutoRecogStr);
}

QString c2bBibParser::setTags(const QString& text)
{
    QString tagged_text = text.trimmed();
    tagged_text.replace(QRegExp("\\r\\n"), "<found_new_line>");       // Windows new line
    tagged_text.replace(QRegExp("\\n"), "<found_new_line>");          // Linux new line, LF
    tagged_text.replace(QRegExp("\\r"), "<found_new_line>");          // OSX new line, CR
    QStringList spText = tagged_text.split("<found_new_line>");
    int n = spText.count();
    tagged_text = "";
    for (int i = 0; i < n - 1; i++)
        tagged_text += spText.at(i) + QString("<NewLine%1>").arg(i + 1);
    tagged_text += spText[n-1];
    spText = tagged_text.split(QRegExp("(\\s{5,}|\\t)"));
    n = spText.count();
    tagged_text = "";
    for (int i = 0; i < n - 1; i++)
        tagged_text += spText.at(i) + QString("<Tab%1>").arg(i + 1);
    tagged_text += spText[n-1];
    tagged_text = tagged_text.simplified();
    return tagged_text;
}

QString c2bBibParser::removeTags(const QString& text)
{
    QString clean = text;
    clean.remove("[[");
    clean.remove("]]");
    clean.replace(QRegExp("<NewLine\\d+>"), " ");
    clean.replace(QRegExp("<Tab\\d+>"), " ");
    clean = clean.simplified();
    return clean;
}

/*! \page heuristic_guess Field Recognition Rules

- <b>Abstract</b>
  - If <tt>Abstract:{0,1}</tt> is found.
  - If <tt>Summary:{0,1}</tt> is found.

*/
void c2bBibParser::guessAbstract(const QString& text)
{
    QRegExp rxH("<NewLine\\d+>\\s*Abstract:{0,1}\\s*<NewLine\\d+>(.+)(<NewLine|$)", Qt::CaseInsensitive);
    rxH.setMinimal(true);
    int nH = rxH.indexIn(text);
    if (nH > -1)
    {
        QString val = rxH.cap(1).remove(QRegExp("^:"));
        bibFields["abstract"]->setText(PProc("abstract", val));
        return;
    }
    rxH = QRegExp("<NewLine\\d+>\\s*Abstract:{0,1}(.+)(<NewLine|$)", Qt::CaseInsensitive);
    rxH.setMinimal(true);
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        QString val = rxH.cap(1).remove(QRegExp("^:"));
        bibFields["abstract"]->setText(PProc("abstract", val));
        return;
    }
    rxH = QRegExp("<NewLine\\d+>\\s*Summary:{0,1}\\s*<NewLine\\d+>(.+)(<NewLine\\d+>\\s*<NewLine){0,1}.+$", Qt::CaseInsensitive);
    rxH.setMinimal(true);
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        QString val = rxH.cap(1).remove(QRegExp("^:"));
        bibFields["abstract"]->setText(PProc("abstract", val));
        return;
    }
    rxH = QRegExp("<NewLine\\d+>\\s*Summary:{0,1}(.+)(<NewLine\\d+>\\s*<NewLine){0,1}.+$", Qt::CaseInsensitive);
    rxH.setMinimal(true);
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        QString val = rxH.cap(1).remove(QRegExp("^:"));
        bibFields["abstract"]->setText(PProc("abstract", val));
    }
}

/*! \page heuristic_guess

- <b>Keywords</b>
  - If <tt>Key\\s{0,1}words:{0,1}</tt> is found.

*/
void c2bBibParser::guessKeywords(const QString& text)
{
    QRegExp rxH("<NewLine\\d+>\\s*Key\\s{0,1}words:{0,1}(.+)(<NewLine|$)", Qt::CaseInsensitive);
    rxH.setMinimal(true);
    int nH = rxH.indexIn(text);
    if (nH > -1)
    {
        QString val = rxH.cap(1).remove(QRegExp("^:"));
        bibFields["keywords"]->setText(PProc("keywords", val));
    }
}

/*! \page heuristic_guess

- <b>Volume</b>
  - If <tt>Volume:{0,1}</tt> is found.
  - If <tt>Vol.{0,1}</tt> is found.
  - If <tt>\\b(\\d+)[,:]\\s*\\d+\\W+\\d+</tt> is found.
  - If <tt>\\b(\\d+)\\s*\\(\\d+\\)</tt> is found.
  - If <tt>\\b(\\d+)[,:]\\s*\\d+\\b</tt> is found.

*/
void c2bBibParser::guessVolume(const QString& text)
{
    QRegExp rxH("Volume:{0,1}\\s*(\\d+)", Qt::CaseInsensitive);
    int nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["volume"]->setText(PProc("volume", rxH.cap(1)));
        return;
    }
    rxH = QRegExp("Vol\\.{0,1}\\s*(\\d+)", Qt::CaseInsensitive);
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["volume"]->setText(PProc("volume", rxH.cap(1)));
        return;
    }
    rxH = QRegExp("\\b(\\d+)[,:]\\s*\\d+\\W+\\d+", Qt::CaseInsensitive);
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["volume"]->setText(PProc("volume", rxH.cap(1)));
        return;
    }
    rxH = QRegExp("\\b(\\d+)\\s*\\(\\d+\\)");
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["volume"]->setText(PProc("volume", rxH.cap(1)));
        return;
    }
    rxH = QRegExp("\\b(\\d+)[,:]\\s*\\d+\\b");
    nH = rxH.indexIn(text);
    if (nH > -1)
        bibFields["volume"]->setText(PProc("volume", rxH.cap(1)));
}

/*! \page heuristic_guess

- <b>Number</b>
  - If <tt>Numbers{0,1}:{0,1}\\s*([\\d-]+)</tt> is found.
  - If <tt>No\\.{0,1}\\s*(\\d+)</tt> is found.
  - If <tt>Issue\\:{0,1}\\s*(\\d+)</tt> is found.
  - If <tt>\\d\\s*\\((\\d+)\\)[^\\.]</tt> is found.

*/
void c2bBibParser::guessNumber(const QString& text)
{
    QRegExp rxH("Numbers{0,1}\\:{0,1}\\s*([\\d-]+)", Qt::CaseInsensitive);
    int nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["number"]->setText(PProc("number", rxH.cap(1)));
        return;
    }
    rxH = QRegExp("No\\.{0,1}\\s*(\\d+)", Qt::CaseInsensitive);
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["number"]->setText(PProc("number", rxH.cap(1)));
        return;
    }
    rxH = QRegExp("Issue\\:{0,1}\\s*(\\d+)", Qt::CaseInsensitive);
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["number"]->setText(PProc("number", rxH.cap(1)));
        return;
    }
    rxH = QRegExp("\\d\\s*\\((\\d+)\\)[^\\.]");
    nH = rxH.indexIn(text);
    if (nH > -1)
        bibFields["number"]->setText(PProc("number", rxH.cap(1)));
}

/*! \page heuristic_guess

- <b>Pages</b>
  - If <tt>\\bPages{0,1}[:\\.]{0,1}([\\d\\s-]+)</tt> is found.
  - If <tt>\\bp{1,2}\\.{0,1}\\s+(\\d+)</tt> is found.
  - If <tt>\\b(\\d+)\\s*-{1,2}\\s*(\\d+)\\b</tt> is found.

*/
void c2bBibParser::guessPages(const QString& text)
{
    QRegExp rxH = QRegExp("\\bPages{0,1}[:\\.]{0,1}([\\d\\s-" + QString(QChar(8211)) + "]+)");
    int nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["pages"]->setText(PProc("pages", rxH.cap(1)));
        return;
    }
    rxH = QRegExp("\\bp{1,2}\\.{0,1}\\s+(\\d+)");
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["pages"]->setText(PProc("pages", rxH.cap(1)));
        return;
    }
    rxH = QRegExp("\\b(\\d+)\\s*[-" + QString(QChar(8211)) + "]{1,2}\\s*(\\d+)\\b");
    nH = 0;
    while (nH >= 0)
    {
        nH = rxH.indexIn(text, nH);
        if (nH > -1)
        {
            if (!rxH.cap(1).startsWith("0"))
            {
                int fp = rxH.cap(1).toInt();
                int lp = rxH.cap(2).toInt();
                if (lp - fp < 150)
                {
                    bibFields["pages"]->setText(PProc("pages", rxH.cap(1) + "-" + rxH.cap(2)));
                    return;
                }
            }
            nH  += rxH.matchedLength();
        }
    }
}

/*! \page heuristic_guess

- <b>Year</b>
  - If <tt>\\b(19|20)(\\d\\d)\\b</tt> is found.

*/
void c2bBibParser::guessYear(const QString& text)
{
    QRegExp rxH("\\b(19|20)(\\d\\d)\\b");
    int nH = rxH.indexIn(text);
    if (nH > -1)
        bibFields["year"]->setText(PProc("year", rxH.cap(1) + rxH.cap(2)));
}

/*! \page heuristic_guess

- <b>Title</b>
  - If <tt>\\bTitle:{0,1}</tt> is found.

*/
void c2bBibParser::guessTitle(const QString& text)
{
    QRegExp rxH("\\bTitle:{0,1}\\s*<NewLine\\d+>(.+)(<NewLine|$)", Qt::CaseInsensitive);
    rxH.setMinimal(true);
    int nH = rxH.indexIn(text);
    if (nH > -1)
    {
        QString val = rxH.cap(1).remove(QRegExp("^:"));
        bibFields["title"]->setText(PProc("title", val));
        return;
    }
    rxH = QRegExp("\\bTitle:{0,1}(.+)(<NewLine|$)", Qt::CaseInsensitive);
    rxH.setMinimal(true);
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        QString val = rxH.cap(1).remove(QRegExp("^:"));
        bibFields["title"]->setText(PProc("title", val));
    }
}

/*! \page heuristic_guess

- <b>ISBN</b>
  - If <tt>\\bISBN\\b(?:-\\d+){0,1}:{0,1}(?:-\\d+){0,1}\\s*(\\d+-[\\d-]+-\\d+)</tt> is found.
  - If <tt>\\bISBN\\b(?:-\\d+){0,1}:{0,1}(?:-\\d+){0,1}\\s*(\\d+)</tt> is found.

*/
void c2bBibParser::guessISBN(const QString& text)
{
    QRegExp rxH("\\bISBN\\b(?:[ -]\\d+){0,1}:{0,1}(?:-\\d+){0,1}\\s*(\\d+-[\\d-]+-\\d+)", Qt::CaseInsensitive);
    int nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["isbn"]->setText(PProc("isbn", rxH.cap(1)));
        return;
    }
    rxH = QRegExp("\\bISBN\\b(?:[ -]\\d+){0,1}:{0,1}(?:-\\d+){0,1}\\s*(\\d+)", Qt::CaseInsensitive);
    nH = rxH.indexIn(text);
    if (nH > -1)
    {
        bibFields["isbn"]->setText(PProc("isbn", rxH.cap(1)));
        return;
    }
}

/*! \page heuristic_guess

- <b>Journal</b>
  - Check cb2Bib internal database.

*/
void c2bBibParser::guessJournal(QString text)
{
    text.replace(QRegExp("\\W"), " ");
    text = text.simplified();
    QString jn = "";

    for (int i = 0; i < JDB->nitems; i++)
    {
        if (text.contains(JDB->JAbbrev_simp_w.at(i), Qt::CaseInsensitive))
        {
            if (jn.length() < JDB->JAbbrev_simp_w.at(i).length())
                jn = JDB->JAbbrev_simp_w.at(i);
        }
    }
    text.replace(QRegExp("\\W"), "");
    for (int i = 0; i < JDB->nitems; i++)
    {
        if (text.contains(JDB->JExtended_simp.at(i), Qt::CaseInsensitive))
        {
            if (jn.length() < JDB->JAbbrev_simp_w.at(i).length())
                jn = JDB->JAbbrev_simp_w.at(i);
        }
    }

    if (!jn.isEmpty())
        bibFields["journal"]->setText(PProc("journal", jn));
}

void c2bBibParser::connectBibWidgets(QHash<QString, c2bLineEdit*>& fields, QComboBox* types, c2bClipEdit* editor)
{
    bibFields = fields;
    bibTypes = types;
    bibEditor = editor;

    // Setting Reference Types
    bibTypes->insertItems(0, TypesList);
    bibTypes->setCurrentIndex(1);

    // Tracking Field Editor Line changes
    QStringList::Iterator it = bibFieldList.begin();
    while (it != bibFieldList.end())
    {
        connect(bibFields[*it], SIGNAL(textChanged(const QString &)), this, SIGNAL(bibModified()));
        it++;
    }
    connect(bibFields["id"], SIGNAL(textChanged(const QString &)), this, SIGNAL(bibModified()));
    connect(bibTypes, SIGNAL(editTextChanged(const QString &)), this, SIGNAL(bibModified()));
    connect(bibFields["journal"], SIGNAL(returnPressed()), this, SLOT(setJournal()));

    // Creating CiteID Maker
    citeIDM = new c2bCiteIDMaker(bibFields, this);
}

int c2bBibParser::countFields()
{
    // Counting Non Empty Fields
    QStringList::Iterator it = bibFieldList.begin();
    int n = 0;
    while (it != bibFieldList.end())
    {
        if (!bibFields[*it]->text().isEmpty())
            n++;
        it++;
    }
    return (n);
}
